"""
Name: EK_AF_UCS_Algorithm.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA
Contact: ryan.j.urbanowicz@darmouth.edu
Created: Sept. 4, 2012
---------------------------------------------------------------------------------------------------------------------------------------------------------
EK_AF_UCS Version 2.0: An expanded learning classifier system algorithm implemented with our new rule compaction strategies for data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRENTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""
#from EK_AF_UCS_Timer import Timer
from EK_AF_UCS_Constants import *
from EK_AF_UCS_ClassifierSet import ClassifierSet
from EK_AF_UCS_Prediction import *
from EK_AF_UCS_AttributeTracking import *
from EK_AF_UCS_RuleCompaction import RuleCompaction
from EK_AF_UCS_ClassAccuracy import ClassAccuracy
from EK_AF_UCS_OutputFileManager import OutputFileManager
import copy
import random

class EK_AF_UCS:
    def __init__(self):
        """ Initializes the EK_AF_UCS algorithm """
        print "EK_AF_UCS: Initializing Algorithm..."
        #Global Parameters-------------------------------------------------------------------------------------
        self.population = None
        self.attributeTracking = None
        self.learnTrackOut = None  #Output file that stores tracking information during learning
        #------------------------------------------------------------------------------------------------------
        if cons.doPopulationReboot: #If we are restarting from a previously saved rule population.
            self.populationReboot()
        else:
            try: #Establish output file to store learning progress.
                self.learnTrackOut = open(cons.outFileName+'_LearnTrack.txt','w')     
                self.learnTrackOut.write("Explore_Iteration\tMacroPopSize\tMicroPopSize\tAccuracy_Estimate\tAveGenerality\tExpRules\tExpWRules\tTime(min)\n")
            except IOError, (errno, strerror):
                print ("I/O error(%s): %s" % (errno, strerror))
                raise
            # Instantiate Population---------
            self.population = ClassifierSet()
            self.exploreIter = 0
            self.correct  = [0.0 for i in range(cons.trackingFrequency)]


    def runEK_AF_UCS(self):
        """ Runs the initialized EK_AF_UCS algorithm. """
        #--------------------------------------------------------------
        print "Learning Checkpoints: " +str(cons.learningCheckpoints)
        print "Maximum Iterations : " +str(cons.maxLearningIterations)
        #--------------------------------------------------------------
        while self.exploreIter < cons.maxLearningIterations: #Major Learning Loop
            
            #Major Algorithm Steps--------------------------------
            state_class = cons.env.getTrainInstance() 
            self.runIteration(state_class, self.exploreIter)
            #----------------------------------------------------------
            
            #EVALUATIONS#############################################################################################################################################################
            cons.timer.startTimeEvaluation()
            #Learning Tracking----------------------------------------------------------------------------------------------------------------------------------------
            if (self.exploreIter%cons.trackingFrequency) == (cons.trackingFrequency - 1) and self.exploreIter > 0:
                cons.completeEpoch() #Assumes that self.trackFreq is set to epoch mode in order to work correctly with RAIN
                self.population.runPopAveEval(self.exploreIter) 
                trackedAccuracy = sum(self.correct)/float(cons.trackingFrequency) #Accuracy over the last "trackingFrequency" number of iterations.
                self.learnTrackOut.write(self.population.getPopTrack(trackedAccuracy, self.exploreIter+1)) #Report learning progress to standard out and tracking file.
            cons.timer.stopTimeEvaluation()
            #---------------------------------------------------------------------------------------------------------------------------------------------------------
            if cons.doArtificialNoise:
                cons.timer.startTimeRAIN()
                self.population.runAttGeneralitySum(False)
                cons.timer.stopTimeRAIN()
                
            #Complete Population Evaluation(s) -----------------------------------------------------------------------------------------------------------------------
            if (self.exploreIter + 1) in cons.learningCheckpoints:
                cons.timer.startTimeEvaluation()
                print "--------------------------------------------------------------------------------------"
                print "Running Complete Population Evaluation at " + str(self.exploreIter + 1)+ " iterations."
                self.population.runPopAveEval(self.exploreIter)
                self.population.runAttGeneralitySum(True)
                cons.env.startEvaluationMode()
                if cons.testFile != 'None': #If a testing file is available.
                    trainEval = self.doPopEvaluation(True)
                    testEval = self.doPopEvaluation(False)
                else:
                    trainEval = self.doPopEvaluation(True)
                    testEval = None
                cons.env.stopEvaluationMode()
                cons.timer.returnGlobalTimer() 
                #Write output files------------------------------------------------------------------------------------------
                OutputFileManager().writePopStats(cons.outFileName, trainEval, testEval, self.exploreIter + 1, self.population, self.correct)
                OutputFileManager().writePop(cons.outFileName, self.exploreIter + 1, self.population)
                if cons.outputAttCoOccur:
                    OutputFileManager().attCo_Occurence(cons.outFileName, self.exploreIter + 1, self.population)
                if cons.doAttributeTracking:
                    OutputFileManager().save_tracking(self.exploreIter, cons.outFileName)
                #------------------------------------------------------------------------------------------------------------
                cons.timer.stopTimeEvaluation()
                print "Continue Learning..."
            ###########################################################################################################################################################################
            
                #RULE COMPACTION ######################################################################################################################################################
                if self.exploreIter + 1 == cons.maxLearningIterations and cons.doRuleCompaction:
                    #---------------------------------------------------------
                    cons.timer.startTimeRuleCmp()
                    RuleCompaction(self.population, trainEval[0], testEval[0])  
                    cons.timer.stopTimeRuleCmp() 
                    #EVALUTAION--------------------------------------------------------------------
                    cons.timer.startTimeEvaluation()
                    self.population.recalculateNumerositySum()
                    self.population.runPopAveEval(self.exploreIter)
                    self.population.runAttGeneralitySum(True)
                    #----------------------------------------------------------
                    cons.env.startEvaluationMode()
                    if cons.testFile != 'None': #If a testing file is available.
                        trainEval = self.doPopEvaluation(True)
                        testEval = self.doPopEvaluation(False)
                    else:
                        trainEval = self.doPopEvaluation(True)
                        testEval = None
                    cons.env.stopEvaluationMode()
                    cons.timer.stopTimeEvaluation()
                    #------------------------------------------------------------------------------
                    cons.timer.returnGlobalTimer() 
                    label = cons.ruleCompactionMethod
                    #Write output files------------------------------------------------------------------------------------------
                    OutputFileManager().writePopStats(cons.outFileName+"_RC_"+str(label), trainEval, testEval, self.exploreIter + 1, self.population, self.correct)
                    OutputFileManager().writePop(cons.outFileName+"_RC_"+str(label), self.exploreIter + 1, self.population)
                    if cons.outputAttCoOccur:
                        OutputFileManager().attCo_Occurence(cons.outFileName+"_RC_"+str(label), self.exploreIter + 1, self.population)
                    #------------------------------------------------------------------------------------------------------------ 
                #######################################################################################################################################################################
                
            self.exploreIter += 1 
            cons.env.newInstance(True) #move to next sample in training set
            
        try: #Close the Learn Tracking File 
            self.learnTrackOut.close()
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise   
        print "EK_AF_UCS Run Complete"
        
        
    def runIteration(self, state_class, exploreIter):
        """ Run single EK_AF_UCS learning iteration. """
        cons.timer.startTimeMatching()
        self.population.makeMatchSet(state_class, exploreIter)
        cons.timer.stopTimeMatching()

        #Prediction for tracking learning progress --------------------------------------
        cons.timer.startTimeEvaluation()
        prediction = Prediction(self.population, len(cons.env.formatData.classList))
        classSelection = prediction.bestClassWinner()
                  
        if classSelection != None:
            if classSelection == state_class[1]:
                self.correct[exploreIter%cons.trackingFrequency] = 1
            else:
                self.correct[exploreIter%cons.trackingFrequency] = 0
        cons.timer.stopTimeEvaluation()
        #--------------------------------------------------------------------------------

        self.population.makeCorrectSet(state_class[1])
        self.population.updateSets(exploreIter)
    
        if cons.doSubsumption:
            cons.timer.startTimeSubsumption()
            self.population.doCorrectSetSubsumption()
            cons.timer.stopTimeSubsumption()
            
        #ATTRIBUTE TRACKING --------------------------------------------
        if cons.doAttributeTracking:
            cons.timer.startTimeAT()
            cons.AT.updateAttTrack(self.population) 
            percent = exploreIter/float(cons.maxLearningIterations) #For timed application of AttTrack
            if cons.doAttributeFeedback and random.random() < percent:
                cons.AT.activateAttFeedback()
                cons.AT.genTrackProb() #Calculates probabilities
            else:
                cons.AT.deactivateAttFeedback()
            cons.timer.stopTimeAT()
        #---------------------------------------------------------------
        
        #GENETIC ALGORITHM------------------------------------------------------------------------------
        self.population.runGA(exploreIter, state_class[0]) #GA is run within the correct set.
        #-----------------------------------------------------------------------------------------------
        self.population.clearSets() #add method to clear sets
        
        
    def doPopEvaluation(self, isTrain):
        """ Performs evaluation of population via the copied environment.  The population is maintained unchanging throughout the evaluation.
        Works on both training and testing data. """
        if isTrain:
            type = "Training"
        else:
            type = "Testing"
        noMatch = 0 #How often does the population fail to have a classifier that matches an instance in the data.
        cons.env.resetDataRef(isTrain) #Go to first instance in data set
        classList = cons.env.formatData.classList #shorter reference to classList - based on training data (assumes no as yet unseen classes in testing data)
        #----------------------------------------------
        classAccDict = {}
        for each in classList:
            classAccDict[each] = ClassAccuracy()
        #----------------------------------------------
        if isTrain:
            samples = cons.env.formatData.numTrainInstances
        else:
            samples = cons.env.formatData.numTestInstances
        #----------------------------------------------------------------------------------------------
        for inst in range(samples):
            if isTrain:
                state_class = cons.env.getTrainInstance()
            else:
                state_class = cons.env.getTestInstance()
            #-----------------------------------------------------------------------------
            self.population.makeEvalMatchSet(state_class[0])
            prediction = Prediction(self.population, len(classList))
            classSelection = prediction.bestClassWinner()

            #-----------------------------------------------------------------------------
            if classSelection == None: 
                noMatch += 1
            else: #Instances which failed to be covered are excluded from the initial accuracy calculation (this is important to the rule compaction algorithm)
                for each in classList:
                    thisIsMe = False
                    accurateClass = False
                    trueClassification = state_class[1]
                    if each == trueClassification:
                        thisIsMe = True #Is the current class the true data class.
                    if classSelection == trueClassification:
                        accurateClass = True
                    classAccDict[each].updateAccuracy(thisIsMe, accurateClass)
            cons.env.newInstance(isTrain) #next instance
            self.population.clearSets() 
        #----------------------------------------------------------------------------------------------
        accSum = 0.0
        print "Class "+str(type)+" Accuracy Summaries:-------------"
        print "-----------------------------------------------"
        
        for each in classList:
            accSum += classAccDict[each].getClassSensitivity() + classAccDict[each].getClassSpecificity()
            #print "Class ("+str(each)+")" #Debug
            #classAccDict[each].reportClassAccuracy() #Debug
        sampleCoverage = 1.0 - (float(noMatch)/float(samples))
        samplesCorrectlyClassified = classAccDict[classList[0]].T_myClass + classAccDict[classList[0]].T_otherClass  

        standardAccuracy = float(samplesCorrectlyClassified) / float(samples - noMatch)
        adjustedStandardAccuracy = (standardAccuracy * sampleCoverage) + ((1.0 - sampleCoverage) * (1.0 / len(classList)))
        
        balancedAccuracy = accSum / (len(classList) * 2.0) #Accuracy Metric for inbalanced and multi-class datasets
        adjustedBalancedAccuracy = (balancedAccuracy * sampleCoverage) + ((1.0 - sampleCoverage) * (1.0 / len(classList))) 

        #Adjusted Balanced Accuracy is calculated such that instances that did not match have a consistent probability of being correctly classified in the reported accuracy.
        print "Overall "+str(type)+" Accuracy Results:-------------"
        print str(samplesCorrectlyClassified) + ' out of ' + str(samples) + ' correctly classified.'
        print "Standard Accuracy = " + str(adjustedStandardAccuracy)
        print "Balanced Accuracy = " + str(adjustedBalancedAccuracy)
        print "Sample Coverage = "+ str(sampleCoverage*100)+ '%'
        print "-----------------------------------------------"
        #Balanced and Standard Accuracies will only be the same when there are equal samples representative of each class AND there is 100% covering. (NOTE even at 100% covering, the values may differ due to subtle float calculation differences in the computer)
        resultList = [adjustedBalancedAccuracy, sampleCoverage]  
        return resultList

    
    def populationReboot(self):
        """ Manages the reformation of a previously saved EK_AF_UCS classifier population. """
        cons.timer.setTimerRestart(cons.popRebootPath) #Rebuild timer objects
        #--------------------------------------------------------------------
        try: #Re-open track learning file for continued tracking of progress.
            self.learnTrackOut = open(cons.outFileName+'_LearnTrack.txt','a')     
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise
        
        #Extract last iteration from file name---------------------------------------------
        temp = cons.popRebootPath.split('_')
        iterRef = len(temp)-1
        completedIterations = int(temp[iterRef])
        print "Rebooting rule population after " +str(completedIterations)+ " iterations."
        self.exploreIter = completedIterations-1
        for i in range(len(cons.learningCheckpoints)):
            cons.learningCheckpoints[i] += completedIterations
        cons.maxLearningIterations += completedIterations

        #Rebuild existing population from text file.--------
        self.population = ClassifierSet(cons.popRebootPath)
        #---------------------------------------------------
        try: #Obtain correct track
            f = open(cons.popRebootPath+"_PopStats.txt", 'r')
            if cons.doAttributeTracking:
                correctRef = 36
            else:
                correctRef = 32
            tempLine = None
            for i in range(correctRef):
                tempLine = f.readline()
            tempList = tempLine.strip().split('\t')
            self.correct = tempList
            for i in range(len(self.correct)):
                self.correct[i] = int(self.correct[i])
            f.close()
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise
        

    def runRConly(self):
        """ Run rule compaction on an existing rule population. """
        print "Initializing Rule Compaction..."
        if not cons.doPopulationReboot:  
            print "Algorithm: Error - Existing population required to run rule compaction alone."
            return
        
        try:
            fileObject = open(cons.popRebootPath+"_PopStats.txt", 'r')  # opens each datafile to read.
        except:
            #break
            print "Data-set Not Found!"
            
        #Retrieve last training and testing accuracies from saved file---------
        tempLine = None
        for i in range(3):
            tempLine = fileObject.readline()
        tempList = tempLine.strip().split('\t')
        trainAcc = float(tempList[0])
        if cons.testFile != 'None': #If a testing file is available.
            testAcc = float(tempList[1])
        else:
            testAcc = None

        #-----------------------------------------------------------------------
        #RULE COMPACTION ######################################################################################################################################################
        cons.timer.startTimeRuleCmp()
        RuleCompaction(self.population, trainAcc, testAcc)  
        cons.timer.stopTimeRuleCmp() 
        #EVALUTAION--------------------------------------------------------------------
        cons.timer.startTimeEvaluation()
        self.population.recalculateNumerositySum()
        self.population.runPopAveEval(self.exploreIter)
        self.population.runAttGeneralitySum(True)
        #----------------------------------------------------------
        cons.env.startEvaluationMode()
        if cons.testFile != 'None': #If a testing file is available.
            trainEval = self.doPopEvaluation(True)
            testEval = self.doPopEvaluation(False)
        else:
            trainEval = self.doPopEvaluation(True)
            testEval = None
        cons.env.stopEvaluationMode()
        cons.timer.stopTimeEvaluation()
        #------------------------------------------------------------------------------
        cons.timer.returnGlobalTimer() 
        #Write output files------------------------------------------------------------------------------------------
        label = cons.ruleCompactionMethod
        OutputFileManager().writePopStats(cons.outFileName+"_RC_"+str(label), trainEval, testEval, self.exploreIter + 1, self.population, self.correct)
        OutputFileManager().writePop(cons.outFileName+"_RC_"+str(label), self.exploreIter + 1, self.population)
        if cons.outputAttCoOccur:
            OutputFileManager().attCo_Occurence(cons.outFileName+"_RC_"+str(label), self.exploreIter + 1, self.population)
        #------------------------------------------------------------------------------------------------------------ 
        print "Rule Compaction Complete"