"""
Name: EK_AF_UCS_Classifier.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA
Contact: ryan.j.urbanowicz@darmouth.edu
Created: Sept. 4, 2012
Description:  This module defines an individual classifier within the rule population, along with all respective parameters.
            Also included are classifier-level methods, including constructors(covering, copy, reboot) matching, subsumption, 
            crossover, and mutation.  Parameter update methods are also included.  
Additional Notes: This module handles classifiers which manage both discrete and continuous attributes.
---------------------------------------------------------------------------------------------------------------------------------------------------------
EK_AF_UCS Version 2.0: An expanded learning classifier system algorithm implemented with our new rule compaction strategies for data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

from EK_AF_UCS_Constants import *
from random import *
import copy
import math

class Classifier:
    def __init__(self,a=None,b=None,c=None):
        """  """
        self.specifiedAttList = [] #Bacardit 2009 - ALKR + GABIL, continuous and discrete rule representation
        self.condition = []   #Bacardit 2009 - ALKR + GABIL, continuous and discrete rule representation
        self.classification = None
        
        #After an epoch the following three values are permanently fixed for the life of the rule
        self.matchCover = 0 #i.e. experience
        self.correctCover = 0  #Rule accuracy is correctCover/matchCover
        self.epochComplete = False  #has this rule existed for a complete epoch.
        
        self.fitness = cons.init_fit
        self.numerosity = 1
        
        #long term-------------------TEMPORARY COMMENT
        self.lastCorrectSetSize = 1
        #OLD WAYS TEST
        self.aveCorrectSetSize = 1
        self.experience = 0
        self.correctTrack = 0
        #New Way Temp
        self.deletionVote = 0
        
        self.timeStampGA = 0 #Time since rule last in a Correct set.
        self.initTimeStamp = 0 #Iteration in which rule first appeared.

        if isinstance(b,list):
            self.classifierCovering(a,b,c)
        elif isinstance(a,Classifier):
            self.classifierCopy(a, b)
        elif isinstance(a,list) and b == None:
            self.rebootClassifier(a)
        else:
            print "Classifier: Error building classifier."
            
            
    #CLASSIFIER CONSTRUCTION METHODS#################################################################################################################################
    def classifierCovering(self, exploreIter, state, classification):
        """ Makes a new classifier when the covering mechanism is triggered.  The new classifier will match the current training instance. """
        #Initialize new classifier parameters----------
        self.lastCorrectSetSize = 1
        self.timeStampGA = exploreIter
        self.initTimeStamp = exploreIter
        #----------------------------------------------
        #Initialize new classifier classification-----
        self.classification = classification
        #----------------------------------------------
        
        #Generate a matching classifier condition----------------------------------------------------
        if cons.useExpertKnowledge:  #Use Expert Knowledge
            for attRef in range(len(state)):
                if random() < cons.EK[attRef] and state[attRef] != cons.labelMissingData: #Specify this attribute in the classifier
                    self.specifiedAttList.append(attRef)
                    self.condition.append(self.buildMatch(attRef, state))

        else: #No Expert Knowledge used
            for attRef in range(len(state)):
                if random() < cons.p_spec and state[attRef] != cons.labelMissingData:
                    self.specifiedAttList.append(attRef)
                    self.condition.append(self.buildMatch(attRef, state))
        #--------------------------------------------------------------------------------------------
            
    def classifierCopy(self, cl, exploreIter):
        """  Constructs an identical Classifier.  However, the experience of the copy is set to 0 and the numerosity 
        is set to 1 since this is indeed a new individual in a population. Used by the genetic algorithm to generate 
        offspring based on parent classifiers."""
        self.specifiedAttList = copy.deepcopy(cl.specifiedAttList)
        self.condition = copy.deepcopy(cl.condition) 
        self.classification = copy.deepcopy(cl.classification)
        self.timeStampGA = exploreIter
        self.initTimeStamp = exploreIter
        
        
    def rebootClassifier(self, classifierList): 
        """ Rebuilds a saved classifier as part of the population Reboot """
        numAttributes = cons.env.formatData.numAttributes
        attInfo = cons.env.formatData.attributeInfo
        for attRef in range(0,numAttributes):
            if classifierList[attRef] != '#':  #Attribute in rule is not wild
                if attInfo[attRef][0]: #Continuous Attribute
                    valueRange = classifierList[attRef].split(cons.alleleSeparation)
                    self.condition.append(valueRange)
                    self.specifiedAttList.append(attRef)
                else:
                    attList = []
                    for j in classifierList[attRef]:
                        attList.append(int(j))
                    self.condition.append(attList)
                    self.specifiedAttList.append(attRef)

        self.classification = str(classifierList[numAttributes])
        self.fitness = float(classifierList[numAttributes+1])
        self.correctCover = int(classifierList[numAttributes+3])
        self.matchCover = float(classifierList[numAttributes+4])
        self.numerosity = int(classifierList[numAttributes+5])
        self.lastCorrectSetSize = float(classifierList[numAttributes+7])
        self.timeStampGA = int(classifierList[numAttributes+8])
        self.initTimeStamp = int(classifierList[numAttributes+9])
        self.epochComplete = bool(int(classifierList[numAttributes+10]))
    ################################################################################################################################################################################
    
    
    def match(self, state):
        """ Returns if the classifier matches in the current situation. """ 
        
        for i in range(len(self.condition)):
            attributeInfo = cons.env.formatData.attributeInfo[self.specifiedAttList[i]]
            if attributeInfo[0]: #Continuous attribute
                instanceValue = state[self.specifiedAttList[i]]
                if self.condition[i][0] < instanceValue < self.condition[i][1] or instanceValue == cons.labelMissingData:
                    pass
                else:
                    return False  
            else: #Discrete attribute
                stateRep = state[self.specifiedAttList[i]]  #only one 1 - as it is the instance encoding
                #print stateRep
                for j in range(len(stateRep)):
                    if stateRep[j] == 0 or (stateRep[j] == 1 and self.condition[i][j] == 1) or stateRep == cons.labelMissingData: #Does the state allele match?
                        pass
                    else:
                        return False    
        return True
        
        
    #GENETIC ALGORITHM###############################################################################################################################################################
    def uniformCrossover(self, cl):
        """ Applies uniform crossover and returns if the classifiers changed. Handles both discrete and continuous attributes. """ 
        #SWARTZ: self. is where for the better attributes are more likely to be specified
        #DEVITO: cl. is where less useful attribute are more likely to be specified
        changed = False;   
        for attRef in range(cons.env.formatData.numAttributes):  #Each condition specifies different attributes, so we need to go through all attributes in the dataset.
            attributeInfo = cons.env.formatData.attributeInfo[attRef]
            if cons.AT.activeAttFeedback: #Attribute Feedback
                probability = cons.AT.getTrackProb()[attRef]
            else:
                probability = 0.5  #Equal probability for attribute alleles to be exchanged.
            #-----------------------------
            ref = 0
            if attRef in self.specifiedAttList:
                ref += 1
            if attRef in cl.specifiedAttList:
                ref += 1
            #-----------------------------
            if ref == 0:    #Attribute not specified in either condition (Attribute type makes no difference)
                pass
            
            elif ref == 1:  #Attribute specified in only one condition - do probabilistic switch of whole attribute state (Attribute type makes no difference)
                if attRef in self.specifiedAttList and random() > probability: # If attribute specified in SWARTZ and high probability of being valuable, then less likely to swap.
                    i = self.specifiedAttList.index(attRef) #reference to the position of the attribute in the rule representation
                    cl.condition.append(self.condition.pop(i)) #Take attribute from self and add to cl
                    cl.specifiedAttList.append(attRef)
                    self.specifiedAttList.remove(attRef)
                    changed = True #Remove att from self and add to cl
                    
                if attRef in cl.specifiedAttList and random() < probability: # If attribute specified in DEVITO and high probability of being valuable, then more likely to swap.
                    i = cl.specifiedAttList.index(attRef) #reference to the position of the attribute in the rule representation
                    self.condition.append(cl.condition.pop(i)) #Take attribute from self and add to cl
                    self.specifiedAttList.append(attRef)
                    cl.specifiedAttList.remove(attRef)
                    changed = True #Remove att from cl and add to self.

            else: #Attribute specified in both conditions - do random crossover between state alleles - Notice: Attribute Feedback must not be used to push alleles together within an attribute state.
                #The same attribute may be specified at different positions within either classifier
                i_cl1 = self.specifiedAttList.index(attRef) #pairs with self (classifier 1)
                i_cl2 = cl.specifiedAttList.index(attRef)   #pairs with cl (classifier 2)
                if attributeInfo[0]: #Continuous attribute 
                    tempKey = randint(0,3) #Make random choice between 4 scenarios, Swap minimums, Swap maximums, Self absorbs cl, or cl absorbs self.
                    if tempKey == 0:    #Swap minimum
                        temp = self.condition[i_cl1][0]
                        self.condition[i_cl1][0] = cl.condition[i_cl2][0]
                        cl.condition[i_cl2][0] = temp
                    elif tempKey == 1:  #Swap maximum
                        temp = self.condition[i_cl1][1]
                        self.condition[i_cl1][1] = cl.condition[i_cl2][1]
                        cl.condition[i_cl2][1] = temp
                    else: #absorb range
                        allList = self.condition[i_cl1] + cl.condition[i_cl2]
                        newMin = min(allList)
                        newMax = max(allList)
                        if tempKey == 2:  #self absorbs cl
                            self.condition[i_cl1] = [newMin,newMax]
                            #Remove cl
                            cl.condition.pop(i_cl2)
                            cl.specifiedAttList.remove(attRef)
                        else:             #cl absorbs self
                            cl.condition[i_cl2] = [newMin,newMax]
                            #Remove self
                            self.condition.pop(i_cl1)
                            self.specifiedAttList.remove(attRef)
  
                else: #Discrete attribute
                    for k in range(len(self.condition[i_cl1])):
                        if random() > 0.5:
                            temp = self.condition[i_cl1][k]
                            self.condition[i_cl1][k]= cl.condition[i_cl2][k]
                            cl.condition[i_cl2][k]= temp
                            changed = True
                    #-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
                    if 0 not in self.condition[i_cl1] or 1 not in self.condition[i_cl1]: #If the attribute encoding is now all 0's or 1's, allele state does not matter so remove it from the classifier.
                        self.condition.pop(i_cl1)
                        self.specifiedAttList.remove(attRef)
                    if 0 not in cl.condition[i_cl2] or 1 not in cl.condition[i_cl2]: #If the attribute encoding is now all 0's or 1's, allele state does not matter so remove it from the classifier.
                        cl.condition.pop(i_cl2)
                        cl.specifiedAttList.remove(attRef)
                    #-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

        return changed


    def mutateCondition(self, state):
        """ Mutates the condition of the classifier.  This is a niche mutation, which means that the resulting classifier will still match the current instance.  """
        changed = False;   
        for attRef in range(cons.env.formatData.numAttributes):  #Each condition specifies different attributes, so we need to go through all attributes in the dataset.
            attributeInfo = cons.env.formatData.attributeInfo[attRef]
            if random() < cons.upsilon and state[attRef] != cons.labelMissingData:
                if cons.AT.activeAttFeedback: #Attribute Feedback
                    probability = cons.AT.getTrackProb()[attRef]
                else: #Normal Mutation Mechanism
                    probability = None
                
                #MUTATION--------------------------------------------------------------------------------------------------------------
                if attRef not in self.specifiedAttList and (not cons.AT.activeAttFeedback or random() < probability): #Attribute not yet specified
                    self.specifiedAttList.append(attRef)
                    self.condition.append(self.buildMatch(attRef, state)) #buildMatch handles both discrete and continuous attributes
                    changed = True
                    
                elif attRef in self.specifiedAttList and (not cons.AT.activeAttFeedback or random() > probability): #Attribute already specified
                    i = self.specifiedAttList.index(attRef) #reference to the position of the attribute in the rule representation
                    
                    if random() > 0.5 or len(self.condition[i]) < 3: #Remove attribute from rule (Attribute type does not matter) - also kicks in if there are only two allele states available (in line with niche mutation).
                        self.specifiedAttList.remove(attRef)
                        self.condition.pop(i) #buildMatch handles both discrete and continuous attributes
                        changed = True
                        
                    else: #Mutate attribute state without removal
                        if attributeInfo[0]: #Continuous attribute 
                            #Mutate continuous range - based on Bacardit 2009 - Select one bound with uniform probability and add or subtract a randomly generated offset to bound, of size between 0 and 50% of att domain.
                            attRange = attributeInfo[1][1] - attributeInfo[1][0]
                            mutateRange = randrange(0.0,0.5)*attRange
                            if random > 0.5: #Mutate minimum 
                                if random > 0.5: #Add
                                    self.condition[i][0] += mutateRange
                                else: #Subtract
                                    self.condition[i][0] -= mutateRange
                            else: #Mutate maximum
                                if random > 0.5: #Add
                                    self.condition[i][1] += mutateRange
                                else: #Subtract
                                    self.condition[i][1] -= mutateRange
                                    
                            #Repair range - such that min specified first, and max second.
                            self.condition[i].sort()
                            changed = True
                            
                        else: #Discrete attribute
                            genList = []
                            numAlleles = len(self.condition[i])
                            codeSum = sum(self.condition[i])
                            if codeSum == 1: #only one allele utilized
                                for k in range(numAlleles):
                                    if self.condition[i][k] == 0:
                                        genList.append(k)
                                self.condition[i][choice(genList)] = 1
                            elif codeSum == (numAlleles - 1): #only one allele not utilized
                                for k in range(numAlleles):
                                    if self.condition[i][k] == 1:
                                        genList.append(k)
                                self.condition[i][choice(genList)] = 1
                            else: # randomly choose whether to add or remove one allele from encoding
                                for k in range(numAlleles):
                                    genList.append(k)
                                theChoice = choice(genList)
                                if self.condition[i][theChoice] == 0:
                                    self.condition[i][theChoice] = 1
                                else:
                                    self.condition[i][theChoice] = 0
                else: #Mutation Canceled
                    pass
                #----------------------------------------------------------------------------------------------------------------------
        return changed
    ############################################################################################################################################################################################################

    def getAccuracy(self):
        """ Calculate and return classifier accuracy. """
        if self.matchCover == 0:
            return 0 
        else:
            return self.correctCover / float(self.matchCover)

        
    #SUBSUMPTION METHODS########################################################################################################################################################################################
    def subsumes(self, cl):
        """ Returns if the classifier (self) subsumes cl. """
        if cl.classification == self.classification:
            if self.isSubsumer() and self.isMoreGeneral(cl):
                return True
        return False

        
    def isSubsumer(self):
        """ Returns if the classifier (self) is a possible subsumer. A classifier must have sufficient experience (one epoch) and it must also be as or more accurate than the classifier it is trying to subsume.  """
        if self.experience > 20  and self.getAccuracy() > 0.99:
            return True
        return False
        #if self.epochComplete and cl.epochComplete:
        #    if self.getAccuracy() >= cl.getAccuracy():
        #        return True
        #return False
    
    
    def isMoreGeneral(self,cl):
        """ Returns if the classifier (self) is more general than cl. Check that all attributes specified in self are also specified in cl. """ 
        if len(self.specifiedAttList) > len(cl.specifiedAttList):
            return False
        
        for i in range(len(self.specifiedAttList)): #Check each attribute specified in self.condition
            attributeInfo = cons.env.formatData.attributeInfo[self.specifiedAttList[i]]
            if self.specifiedAttList[i] not in cl.specifiedAttList:
                return False

            otherRef = cl.specifiedAttList.index(self.specifiedAttList[i])
            if attributeInfo[0]: #Continuous attribute
                #If self has a narrower ranger of values than it is a subsumer
                if self.condition[i][0] < cl.condition[otherRef][0]:
                    return False
                if self.condition[i][1] > cl.condition[otherRef][1]:
                    return False
                
            else: #Discrete attribute
                #If cl specifies every allele that self does, than self is a subsumer.
                if sum(self.condition[i]) > sum(cl.condition[otherRef]):
                    return False

                for k in range(len(self.condition[i])):
                    if self.condition[i][k] == 1 and cl.condition[otherRef][k] != 1:
                        return False
        return True
    
    ############################################################################################################################################################################################################
    
    #DELECTION##################################################################################################################################################################################################
    def getDeletionVote(self, exploreIter, meanFitness):
        """  Returns the vote for deletion of the classifier. """
        deletionVote = 1.0
        if self.numerosity > 1 and self.numerosity >= 0.01 * cons.N: #for N = 1000, won't kick in until numerosity = 10
            deletionVote += math.exp(self.numerosity / (0.01 * cons.N))
              
        if self.lastCorrectSetSize > 0.1 * cons.N:
            deletionVote += 1.0 
            
        epochProgress = ((exploreIter - self.initTimeStamp) / float(cons.env.formatData.numTrainInstances))
        if (self.epochComplete and self.getAccuracy() <= 0.5) or (not self.epochComplete and self.fitness != 0.0 and self.getAccuracy() <= 0.5 and random() < epochProgress): # Huge Probability of deletion.
            deletionVote = deletionVote / float(cons.init_fit)
        else:
            deletionVote += 1.0 - self.getAccuracy() # NOTE: fitness metric needs to be on scale of 0-1

        self.deletionVote = deletionVote
        return self.deletionVote
    ############################################################################################################################################################################################################
        """ #FUNCTIONAL
        deletionVote = 1.0
        if self.numerosity > 1 and self.numerosity >= 0.01 * cons.N: #for N = 1000, won't kick in until numerosity = 10
            deletionVote += self.numerosity / (0.01 * cons.N)
              
        if self.lastCorrectSetSize > 0.1 * cons.N:
            deletionVote += 1.0 
            
        epochProgress = ((exploreIter - self.initTimeStamp) / float(cons.env.formatData.numTrainInstances))
        if (self.epochComplete and self.getAccuracy() <= 0.5) or (not self.epochComplete and self.fitness != 0.0 and self.getAccuracy() <= 0.5 and random() < epochProgress): # Huge Probability of deletion.
            deletionVote = deletionVote / float(cons.init_fit)
        else:
            deletionVote += 1.0 - self.getAccuracy() # NOTE: fitness metric needs to be on scale of 0-1

        self.deletionVote = deletionVote
        return self.deletionVote
        """
        
    def buildMatch(self, attRef, state):
        """ Builds a matching condition for the classifierCovering method. """
        attributeInfo = cons.env.formatData.attributeInfo[attRef]
        if attributeInfo[0]:  #If continuous attribute
            attRange = attributeInfo[1][1] - attributeInfo[1][0]
            rangeRadius = randint(25,75)*0.01*attRange / 2.0 #Continuous initialization domain radius.
            Low = state[attRef] - rangeRadius
            High = state[attRef] + rangeRadius
            condList = [Low,High] #ALKR Representation, Initialization centered around training instance  with a range between 25 and 75% of the domain size.
        else: #If discrete attribute
            condList = state[attRef] #State already formatted like GABIL in DataManagement
            
        return condList
     

    def equals(self, cl):  
        """ Returns if the two classifiers are identical in condition and action. """ 
        if cl.classification == self.classification: #Is classification the same
            if cl.condition == self.condition: #Is a rule the same  
                return True
        return False


    #UPDATES####################################################################################################################################################################################################  
    def updateEpochStatus(self, exploreIter):
        """ Determines when a learning epoch has completed (one cycle through training data). """
        if not self.epochComplete and (exploreIter - self.initTimeStamp) >= cons.env.formatData.numTrainInstances:
            self.epochComplete = True
        
            
    def updateFitness(self, exploreIter):
        """ Update the fitness parameter. """ #get rid of exploreIter if not being used.
        """
        if self.epochComplete:
            #pass
            accuracy = self.getAccuracy()
            if accuracy > 0.5:
                self.fitness = math.pow((accuracy - 0.5), cons.nu) * self.correctCover 
            else:
                self.fitness = accuracy
        else:
            #self.fitness = pow(self.getAccuracy(), cons.nu)
            epochProgress = ((exploreIter - self.initTimeStamp + 1) / float(cons.env.formatData.numTrainInstances))
            accuracy = self.getAccuracy()
            if accuracy > 0.5:
                self.fitness = math.pow((accuracy - 0.5), cons.nu) * self.correctCover / epochProgress
            else:
                self.fitness = accuracy
            #self.fitness = self.fitness / float(cons.formatData.numTrainInstances)
            #self.fitness = (accuracy - 0.5) * self.correctCover / epochProgress
        """
        """
        accuracy = self.getAccuracy()
        if accuracy > 0.5:
            self.fitness = math.pow((accuracy - 0.5), cons.nu) * self.correctCover 
        else:
            self.fitness = accuracy
        """
        """
        if self.epochComplete:
            pass
        else:
            accuracy = self.getAccuracy()
            if random() > 0.5:
                self.fitness = pow(accuracy, cons.nu)
            else:
                if accuracy > 0.5:
                    self.fitness = math.pow((accuracy - 0.5), cons.nu) * self.correctCover 
                else:
                    self.fitness = accuracy
        """
        

        if self.epochComplete:
            pass
        else:
            self.fitness = pow(self.getAccuracy(), cons.nu)

        
        
    def updateExperience(self):
        """ Increases the experience of the classifier by one. Once an epoch has completed, rule accuracy can't change."""
        self.experience += 1 #DEBUG
        if self.epochComplete:
            pass
        else:
            self.matchCover += 1


    def updateCorrect(self):
        """ Increases the correct classification tracking by one. Once an epoch has completed, rule accuracy can't change."""
        self.correctTrack += 1 #DEBUG
        if self.epochComplete:
            pass
        else:
            self.correctCover += 1
        
        
    #The following updates will continue to change even after the first epoch-----------------------------------------------------------------------------------------------------------------------------------
    def updateNumerosity(self, num):
        """ Alters the numberosity of the classifier.  Notice that num can be negative! """
        self.numerosity += num
        
        
    def updateCorrectSetSize(self, correctSetSize):
        """  Updates the correct set size. """
        self.lastCorrectSetSize = correctSetSize
        #self.correctExperience += 1
        #OLD correct size
        if self.experience < 1.0 / 0.1:
            self.aveCorrectSetSize = (self.aveCorrectSetSize * (self.experience-1)+ correctSetSize) / float(self.experience)
        else:
            self.aveCorrectSetSize = self.aveCorrectSetSize + 0.1 * (correctSetSize - self.aveCorrectSetSize)

        return self.aveCorrectSetSize * self.numerosity
        
        
    def updateTimeStamp(self, ts):
        """ Sets the time stamp of the classifier. """
        self.timeStampGA = ts
        
        
    ############################################################################################################################################################################################################
    def printClassifier(self):
        """ Formats and returns an output string describing this classifier. """ 
        classifierString = ""
        for attRef in range(cons.env.formatData.numAttributes):
            attributeInfo = cons.env.formatData.attributeInfo[attRef]
            if attRef in self.specifiedAttList:  #If the attribute was specified in the rule
                i = self.specifiedAttList.index(attRef)
                if attributeInfo[0]: #Continuous attributes
                    classifierString += str(self.condition[i][0])+str(cons.alleleSeparation)+str(self.condition[i][1]) + "\t"
                else: #Discrete attributes
                    alleleString = ''
                    for j in self.condition[i]:
                        alleleString += str(j)
                    classifierString += str(alleleString) + "\t"
            else: # Attribute is wild.
                classifierString += '#' + "\t"
        #-------------------------------------------------------------------------------
        specificity = len(self.condition) / float(cons.env.formatData.numAttributes)
        
        epoch = 0
        if self.epochComplete:
            epoch = 1
        #------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
        classifierString += str(self.classification)+"\t"+str(self.fitness)+"\t"+str(self.getAccuracy())+"\t"+str(self.correctCover)+"\t"+str(self.matchCover)+"\t"+str(self.numerosity)+"\t"+str(specificity)+"\t"
        classifierString += str(self.lastCorrectSetSize)+"\t"+str(self.timeStampGA)+"\t"+str(self.initTimeStamp)+"\t"+str(epoch)+"\t"+str(self.deletionVote)+"\n"

        #------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
        return classifierString