"""
Name: EK_AF_UCS_ClassifierSet.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA
Contact: ryan.j.urbanowicz@darmouth.edu
Created: Sept. 4, 2012
---------------------------------------------------------------------------------------------------------------------------------------------------------
EK_AF_UCS Version 2.0: An expanded learning classifier system algorithm implemented with our new rule compaction strategies for data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

from EK_AF_UCS_Constants import *
from EK_AF_UCS_Classifier import Classifier
from random import *
import copy
import sys

class ClassifierSet:
    def __init__(self, a=None):
        """ Overloaded initialization: Handles creation of different sets required by algorithm. (i.e. Population, match set, action set) """
        self.popSet = [] #list of classifiers/rules
        self.matchSet = [] #list of references to rules in population that match
        self.correctSet = [] #list of references to rules in population that both match and specify correct classification
        self.numerositySum = 0
        
        #Evaluation Parameters-------------------------------
        self.aveGenerality = 0.0
        self.expRules = 0.0
        self.expRulesWeighted = 0.0
        self.attributeSpecList = []
        self.attributeAccList = []
        
        #Set Constructors-------------------------------------
        if a==None:
            self.makePop() 
        elif isinstance(a,str):
            self.rebootPop(a)
        else:
            print "ClassifierSet: Error building population."
            

    #CLASSIFIER SET CONSTRUCTION METHODS#################################################################################################################################
    def makePop(self):
        """ Initializes the rule population """
        self.popSet = []
            
            
    def rebootPop(self, remakeFile):
        """ Remakes a previously evolved population from a saved text file. """
        print "Rebooting the following population: " + str(remakeFile)+"_RulePop.txt"
        #*******************Initial file handling**********************************************************
        try:       
            datasetList = []
            f = open(remakeFile+"_RulePop.txt", 'r')
            self.headerList = f.readline().rstrip('\n').split('\t')   #strip off first row
            attType = f.readline().rstrip('\n').split('\t')
            attCode = f.readline().rstrip('\n').split('\t')
            for line in f:
                lineList = line.strip('\n').split('\t')
                datasetList.append(lineList)
            f.close()
            
        except IOError, (errno, strerror):
            print ("Could not Read Remake File!")
            print ("I/O error(%s): %s" % (errno, strerror))
            raise
        except ValueError:
            print ("Could not convert data to an integer.")
            raise
        except:
            print ("Unexpected error:", sys.exc_info()[0])
            raise
        #**************************************************************************************************
        for each in datasetList:
            cl = Classifier(each)
            self.popSet.append(cl) #Add classifier to the population
            numerosityRef = cons.env.formatData.numAttributes + 5
            self.numerositySum += int(each[numerosityRef])

    
    def makeMatchSet(self, state_class, exploreIter):
        """ Constructs a match set from the population. Covering is initiated if the match set is empty or a rule with the current correct class is absent. 
         If maximal population size is reached when covering, deletion occurs. """ 
        state = state_class[0]
        classification = state_class[1]
        doCovering = True # Covering check: Twofold (1)checks that a match is present, and (2) that at least one match dictates the correct class.
        
        #Matching----------------------------------------------------------------------
        cons.timer.startTimeMatching()                          #TIME TRACK
        for i in range(len(self.popSet)):                       #Go through the population
            cl = self.popSet[i]                                 #A single classifier
            cl.updateEpochStatus(exploreIter)  #Note whether this classifier has seen all training data by now.
            if cl.match(state):                                 #Check for match
                self.matchSet.append(i)                         #Add classifier to match set
                #Covering Check--------------------------------------------------------         
                if cl.classification == classification: #Check that class is correct
                    doCovering = False
                #----------------------------------------------------------------------
        cons.timer.stopTimeMatching()                   #TIME TRACK
        #------------------------------------------------------------------------------
        
        #Covering----------------------------------------------------------------------
        while doCovering:
            newCl = Classifier(exploreIter, state, classification)
            self.addClassifierToPopulation(newCl)
            self.matchSet.append(len(self.popSet)-1)  #add covered classifier to matchset
            doCovering = False
        #------------------------------------------------------------------------------
                
            #Deletion----------------------------------------------------------------------  
            cons.timer.startTimeDeletion()                  #TIME TRACK
            while self.numerositySum > cons.N:  
                clInfo = self.deleteFromPopulation(exploreIter) #Delete classifier from population set
                if clInfo != None and clInfo[0] in self.matchSet and clInfo[1] == classification: #Checks if a rule was selected for deletion and whether it is in the match set and if that deleted rule had the correct classification.
                    doCovering = True
            cons.timer.stopTimeDeletion()                   #TIME TRACK
            #------------------------------------------------------------------------------

        
    def makeCorrectSet(self, classification):
        """ Constructs a correct set out of the given match set. """      
        for i in range(len(self.matchSet)):
            ref = self.matchSet[i]
            if self.popSet[ref].classification == classification:
                self.correctSet.append(ref) #HOME TO THE BIGGEST MFSOB BUG EVER - YOU TIME WASTING SOB!

                
    def makeEvalMatchSet(self, state):
        """ Constructs a match set which does not include covering or deletion, so as not to disrupt the existing population
        and allow for it's evaluation in a static manner as an evolved solution. """
        for i in range(len(self.popSet)):               #Go through the population
            cl = self.popSet[i]                         #A single classifier
            if cl.match(state):                         #Check for match
                self.matchSet.append(i)                 #Add classifier to match set    
                  
    #################################################################################################################################################################### 
    
    def addClassifierToPopulation(self, cl):
        """ Adds a classifier to the set and increases the numerositySum value accordingly."""
        self.popSet.append(cl)
        self.numerositySum += cl.numerosity
        
        
    #CLASSIFIER DELETION METHODS#########################################################################################################################################
    def deleteFromPopulation(self, exploreIter):
        """ Deletes one classifier in the population.  The classifier that will be deleted is chosen by roulette wheel selection
        considering the deletion vote. Returns the macro-classifier which got decreased by one micro-classifier. """
        meanFitness = self.getPopFitnessSum()/float(self.numerositySum)
        #Calculate total wheel size------------------------------
        sumCl = 0.0
        voteList = []
        for cl in self.popSet:
            vote = cl.getDeletionVote(exploreIter, meanFitness)
            sumCl += vote
            voteList.append(vote)
        #--------------------------------------------------------
        choicePoint = sumCl * random() #Determine the choice point

        newSum=0.0
        for i in range(len(voteList)):
            cl = self.popSet[i]
            newSum = newSum + voteList[i]
            if newSum > choicePoint: #Select classifier for deletion
                #Delete classifier----------------------------------
                clInfo = [i,cl.classification]
                cl.updateNumerosity(-1)
                self.numerositySum -= 1
                if cl.numerosity < 1: # When all micro-classifiers for a given classifier have been depleted.
                    self.removeMacroClassifier(i)
                    self.deleteFromMatchSet(i)  #found error
                    self.deleteFromCorrectSet(i)
                return clInfo

        print "ClassifierSet: No eligible rules found for deletion in deleteFrom population."
        return None
        #if rule is not in match set - still need to adjust match set

    def removeMacroClassifier(self, ref):
        """ Removes the specified (possible macro-) classifier from the population.
        The function returns true when the classifier was found and removed and false
        otherwise. It does not update the numerosity sum of the set, neither
        recursively remove classifiers in the parent set. """
        try:
            self.popSet.pop(ref)
            return True
        except ValueError:
            raise
        
        
    def deleteFromMatchSet(self, deleteRef):
        """ Delete reference to classifier in population, contained in self.matchSet."""
        if deleteRef in self.matchSet:
            self.matchSet.remove(deleteRef)
        #Update match set reference list--------
        for j in range(len(self.matchSet)):
            ref = self.matchSet[j]
            if ref > deleteRef:
                self.matchSet[j] -= 1
        
        
    def deleteFromCorrectSet(self, deleteRef):
        """ Delete reference to classifier in population, contained in self.matchSet."""
        if deleteRef in self.correctSet:
            self.correctSet.remove(deleteRef)
        #Update match set reference list--------
        for j in range(len(self.correctSet)):
            ref = self.correctSet[j]
            if ref > deleteRef:
                self.correctSet[j] -= 1
            
    #################################################################################################################################################################### 

    def runGA(self, exploreIter, state):
        """ The genetic discovery mechanism in EK_AF_UCS is controlled here. """
        if len(self.correctSet) < 1 or (exploreIter - self.getIterStampAverage()) < cons.theta_GA:  #Does the correct set meet the requirements for activating the GA?
            return 
        
        self.setIterStamps(exploreIter) #Show that all included rules have been involved with GA in current iteration.
        changed = False
        
        #Selection Method---------------------------------------------------------------
        cons.timer.startTimeSelection()
        if cons.selectionMethod == "roulette": 
            selectList = self.selectClassifierRW()
            clP1 = selectList[1]
            clP2 = selectList[0]
        elif cons.selectionMethod == "tournament":
            selectList = self.selectClassifierT()
            clP1 = selectList[1]
            clP2 = selectList[0]
        else:
            print "ClassifierSet: Error - requested GA selection method not available."
        cons.timer.stopTimeSelection()
        #-------------------------------------------------------------------------------
        
        # Create offspring copies of the parents which have 0 experience and 1 numerosity.
        cl1  = Classifier(clP1, exploreIter)
        if clP2 == None:
            cl2 = Classifier(clP1, exploreIter)
        else:
            cl2 = Classifier(clP2, exploreIter)
            
        #Crossover----------------------------------------------------------------------
        if clP2 != None and random() < cons.chi:  
            cons.timer.startTimeCrossover()
            if cons.crossoverMethod == "spc":
                changed = cl1.singlePointCrossover(cl2)
            elif cons.crossoverMethod == "dpc":
                changed = cl1.twoPointCrossover(cl2)
            elif cons.crossoverMethod == "uc":
                changed = cl1.uniformCrossover(cl2)
            else:
                print "ClassifierSet: Error - requested crossover method not available!"
            cons.timer.stopTimeCrossover()
        #-------------------------------------------------------------------------------

        #Mutation-----------------------------------------------------------------------
        cons.timer.startTimeMutation()
        nowchanged = cl1.mutateCondition(state)
        howaboutnow = cl2.mutateCondition(state)
        cons.timer.stopTimeMutation()
        #-------------------------------------------------------------------------------
        if changed or nowchanged or howaboutnow:
            self.insertDiscoveredClassifiers(cl1, cl2, clP1, clP2, exploreIter) #Subsumption
            

    def getIterStampAverage(self):
        """ Returns the average of the time stamps in the set. """
        sumCl=0.0
        for cl in self.popSet:
            sumCl += cl.timeStampGA * cl.numerosity
        return sumCl/self.numerositySum
    

    def setIterStamps(self, exploreIter):
        """ Sets the time stamp of all classifiers in the set to the current time. The current time
        is the number of exploration steps executed so far.  """
        for i in range(len(self.correctSet)):
            ref = self.correctSet[i]
            self.popSet[ref].updateTimeStamp(exploreIter)


    def getFitnessSum(self, setList):
        """ Returns the sum of the fitnesses of all classifiers in the set. """
        sumCl=0.0
        for i in range(len(setList)):
            ref = setList[i]
            sumCl += self.popSet[ref].fitness
        return sumCl


    def getPopFitnessSum(self):
        """ Returns the sum of the fitnesses of all classifiers in the set. """
        sumCl=0.0
        for cl in self.popSet:
            sumCl += cl.fitness * cl.numerosity
        return sumCl
    
    
    #SELECTION METHODS###########################################################################################################################################################
    def selectClassifierRW(self):
        """ Selects parents using roulette wheel selection according to the fitness of the classifiers. """
        selectList = [None, None]
        setList = copy(self.correctSet)
        if len(setList < 2):
            currentCount = 1 #Pick only one parent
        else:
            currentCount = 0 #Pick two parents
        #-----------------------------------------------
        while currentCount < 2:
            fitSum = self.getFitnessSum(setList)
            
            choiceP = random() * fitSum
            i=0
            ref = self.setList[i]
            sumCl = self.popSet[ref].fitness
            while choiceP > sumCl:
                i=i+1
                sumCl += self.popSet[ref].fitness
                
            selectList[currentCount] = self.popSet[ref]
            setList.remove(ref)
            currentCount += 1
        #-----------------------------------------------
        return selectList


    def selectClassifierT(self):
        """  Selects parents using tournament selection according to the fitness of the classifiers. """
        selectList = [None, None]
        setList = copy.copy(self.correctSet)
        if len(setList) < 2:
            currentCount = 1 #Pick only one parent
        else:
            currentCount = 0 #Pick two parents
        #-----------------------------------------------
        while currentCount < 2:
            tSize = int(round(len(setList) * cons.theta_sel)) # determines the number of classifiers in the correct set to be included in the tournament selection
            posList = sample(setList,tSize)    #references to classifiers
            bestFit = 0
            bestCl = posList[0] #reference
            for ref in posList:
                if self.popSet[ref].fitness > bestFit:
                    bestFit = self.popSet[ref].fitness
                    bestCl = ref
                    
            selectList[currentCount] = self.popSet[bestCl]
            setList.remove(bestCl)
            currentCount += 1
        #-----------------------------------------------

        return selectList # returning references to rule(s) in population
    

    ##################################################################################################################################################################################     
    def insertDiscoveredClassifiers(self, cl1, cl2, clP1, clP2, exploreIter):
        """ Inserts both discovered classifiers keeping the maximal size of the population and possibly doing GA subsumption. """
        #Subsumption ------------------------------------------
        if cons.doSubsumption:
            cons.timer.startTimeSubsumption()
            self.subsumeClassifier(cl1, clP1, clP2)
            self.subsumeClassifier(cl2, clP1, clP2)
            cons.timer.stopTimeSubsumption()
        #------------------------------------------------------
        else: #Just add the new classifiers to the population.
            self.addClassifierToPopulation(cl1)
            self.addClassifierToPopulation(cl2)
        
        #DELETION - Attempts to maintain max population size-------------
        cons.timer.startTimeDeletion()
        while self.numerositySum > cons.N:
            self.deleteFromPopulation(exploreIter)
        cons.timer.stopTimeDeletion()
        #----------------------------------------------------------------
        
        
    #SUBSUMPTION METHODS##############################################################################################################################################################
    def subsumeClassifier(self, cl=None, cl1P=None, cl2P=None):
        """ Tries to subsume a classifier in the parents. If no subsumption is possible it tries to subsume it in the current set. """
        if cl1P!=None and cl1P.subsumes(cl):
            self.numerositySum += 1
            cl1P.updateNumerosity(1)
        elif cl2P!=None and cl2P.subsumes(cl):
            self.numerositySum += 1
            cl2P.updateNumerosity(1)
        else:
            self.subsumeClassifier2(cl); #Try to subsume in the correct set.


    def subsumeClassifier2(self, cl):
        """ Tries to subsume a classifier in the correct set. If no subsumption is possible the classifier is simply added to the population considering
        the possibility that there exists an identical classifier. """
        choices = []
        for ref in self.correctSet:
            if self.popSet[ref].subsumes(cl):
                choices.append(ref)

        if len(choices) > 0: #Randomly pick one classifier to be subsumer
            choice = int(random()*len(choices))
            self.popSet[choices[choice]].updateNumerosity(1)
            self.numerositySum += 1
            return
        
        #If no subsumer was found, check for identical classifier, if not then add the classifier to the population
        self.getIdenticalClassifier(cl)
        #----------------------------------------------------------------------------------------------------------


    def doCorrectSetSubsumption(self):
        """ Executes correct set subsumption.  The correct set subsumption looks for the most general subsumer classifier in the correct set
        and subsumes all classifiers that are more specific than the selected one. """
        subsumer = None
        for ref in self.correctSet:
            cl = self.popSet[ref]
            if subsumer == None or cl.isMoreGeneral(subsumer):
                subsumer = cl

        if subsumer != None: #If a subsumer was found, subsume all more specific classifiers in the correct set
            i=0
            while i < len(self.correctSet):
                ref = self.correctSet[i]
                if subsumer != self.popSet[ref] and subsumer.subsumes(self.popSet[ref]): 
                    subsumer.updateNumerosity(self.popSet[ref].numerosity)
                    self.removeMacroClassifier(ref)
                    self.deleteFromMatchSet(ref)
                    self.deleteFromCorrectSet(ref)
                    i = i - 1 
                i = i + 1
                
                
    ################################################################################################################################################################################
    def getIdenticalClassifier(self, newCl):
        """ Looks for an identical classifier in the population. """
        oldCl = None
        for cl in self.popSet:
            if newCl.equals(cl):
                oldCl = cl

        if oldCl != None: 
            oldCl.updateNumerosity(1)
            self.numerositySum += 1
        else:
            self.addClassifierToPopulation(newCl) 
            
             
    def clearSets(self):
        """ Clears out references in the match and correct sets for the next learning iteration. """
        self.matchSet = []
        self.correctSet = []
        

    def updateSets(self, exploreIter):
        """ Updates all relevant parameters in the current match and correct sets. """
        for ref in self.matchSet:
            self.popSet[ref].updateExperience()    
            
            if ref in self.correctSet:
                self.popSet[ref].updateCorrect()
                self.popSet[ref].updateCorrectSetSize(len(self.correctSet))
            self.popSet[ref].updateFitness(exploreIter)
        

    #EVALUTATION METHODS###############################################################################################################################################################################
    def runPopAveEval(self, exploreIter):
        """ Determines current generality of population """
        genSum = 0
        agedCount = 0
        agedCountWeighted = 0
        for cl in self.popSet:
            genSum += ((cons.env.formatData.numAttributes - len(cl.condition)) / float(cons.env.formatData.numAttributes)) * cl.numerosity
            if (exploreIter - cl.initTimeStamp) > cons.env.formatData.numTrainInstances:
                agedCount += 1
                agedCountWeighted += cl.numerosity
                
        if self.numerositySum == 0:
            self.aveGenerality = 'NA'
            self.expRules = 'NA'
            self.expRulesWeighted = 'NA'
        else:
            self.aveGenerality = genSum / float(self.numerositySum) 
            self.expRules = agedCount / float(len(self.popSet))
            self.expRulesWeighted = agedCountWeighted / float(self.numerositySum)
        
    def runAttGeneralitySum(self, isEvaluationSummary):
        """ Determine the population-wide frequency of attribute specification, and accuracy weighted specification. Used for population evaluation as well as by RAIN if activated. """
        if isEvaluationSummary:
            self.attributeSpecList = []
            self.attributeAccList = []
            for i in range(cons.env.formatData.numAttributes):
                self.attributeSpecList.append(0)
                self.attributeAccList.append(0.0)
            for cl in self.popSet:
                for ref in cl.specifiedAttList: #for each attRef
                    #Numerosity taken into account
                    self.attributeSpecList[ref] += cl.numerosity
                    self.attributeAccList[ref] += cl.numerosity * cl.getAccuracy()

        else: # For RAIN: Every epoch calculate desired weight values.
            weightList = []
            for i in range(cons.env.formatData.numAttributes):
                weightList.append(0)
            for cl in self.popSet:
                for ref in cl.specifiedAttList: #for each attRef
                    #Numerosity taken into account
                    if cons.noiseWeightType == 'Spec':
                        weightList[ref] += cl.numerosity
                    elif cons.noiseWeightType == 'AWSpec':
                        weightList[ref] += cl.numerosity * (cl.correctCover / cl.matchCover)
                    elif cons.noiseWeightType == 'FWSpec':
                        weightList[ref] += cl.fitness
                    else:
                        print "Constants: Error - RAIN weighting scheme not found."
            cons.updateRAINWeights(weightList) #Send weights to constants so that makeItRain can access them.
  

    def recalculateNumerositySum(self):
        """ Recalculate the NumerositySum after rule compaction. """
        self.numerositySum = 0
        for cl in self.popSet:
            self.numerositySum += + cl.numerosity
              
    ####################################################################################################################################################################################################
    
    def getPopTrack(self, accuracy, exploreIter):
        """ Returns a formated output string to be printed to the Learn Track output file. """
        trackString = str(exploreIter)+ "\t" + str(len(self.popSet)) + "\t" + str(self.numerositySum) + "\t" + str(accuracy) + "\t" + str(self.aveGenerality) + "\t" + str(self.expRules) + "\t" + str(self.expRulesWeighted) + "\t" + str(cons.timer.returnGlobalTimer())+ "\n"
        print ("Iteration: " + str(exploreIter) + "\t MacroPop: " + str(len(self.popSet))+ "\t MicroPop: " + str(self.numerositySum) + "\t AccEstimate: " + str(accuracy) + "\t AveGen: " + str(self.aveGenerality) + "\t ExpRules: " + str(self.expRules) + "\t ExpWRules: " + str(self.expRulesWeighted) + "\t Time: " + str(cons.timer.returnGlobalTimer()))
        return trackString
        
    ####################################################################################################################################################################################################  