"""
Name: EK_AF_UCS_DataManagement.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA
Contact: ryan.j.urbanowicz@darmouth.edu
Created: Sept. 4, 2012
Description: Able to manage both training and testing data.  This module loads the dataset, detects and characterizes all attributes in the dataset, 
            handles missing data, and finally formats the data so that it may be conveniently utilized by EK_AF_UCS.
Additional Notes: Includes some Rule-representation specific code.  Assumes that testing and training datasets are formatted/organized in the same way.
---------------------------------------------------------------------------------------------------------------------------------------------------------
EK_AF_UCS Version 2.0: An expanded learning classifier system algorithm implemented with our new rule compaction strategies for data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

from EK_AF_UCS_Constants import * #Constants Module referenced by 'cons'
from random import shuffle
import sys

class DataManagement:
    def __init__(self, trainFile, testFile):
        #Initialize global variables-------------------------------------------------
        self.numAttributes = None  # Saves the number of attributes in the input file. 
        self.areInstanceIDs = False
        self.instanceIDRef = None
        self.classRef = None

        self.attributeInfo = [] #Stores Discrete (0) vs. Continuous (1)
        self.classList = [] #cases and controls
        
        #Train/Test Specific-----------------------------------------------------------------------------
        self.trainHeaderList = []
        self.testHeaderList = []
        self.numTrainInstances = None 
        self.numTestInstances = None 
        #------------------------------------------------------------------------------------------------
        
        #Detect and format training data--------------------------------------------------------------------------
        print "----------------------------------------------------------------------------"
        print "Environment: Formatting Training Data... "
        rawTrainData = self.loadData(trainFile, True)   #Load the raw data.
        self.characterizeDataset(rawTrainData)
        self.discriminateClasses(rawTrainData)
        self.discriminateAttributes(rawTrainData)
        
        if cons.testFile == 'None':
            data4Formating = rawTrainData
        else:
            print "----------------------------------------------------------------------------"
            print "Environment: Formatting Testing Data... "
            rawTestData = self.loadData(testFile, False)
            self.compareDataset(rawTestData)
            data4Formating = rawTrainData + rawTestData
            
        self.characterizeAttributes(data4Formating)
        
        self.trainFormatted = self.formatData(rawTrainData) #Stores the formatted training data set used throughout the algorithm.
        if cons.testFile != 'None':
            self.testFormatted = self.formatData(rawTestData) #Stores the formatted testing data set used throughout the algorithm.
        print "----------------------------------------------------------------------------"
        
            
    def loadData(self, dataFile, doTrain):
        """ Load the data file. """     
        #************************************************************************************************* 
        print "DataManagement: Loading Data... " + str(dataFile)
        try:       
            datasetList = []
            f = open(dataFile,'r')
            if doTrain:
                self.trainHeaderList = f.readline().rstrip('\n').split('\t')   #strip off first row
            else:
                self.testHeaderList = f.readline().rstrip('\n').split('\t')   #strip off first row
            for line in f:
                lineList = line.strip('\n').split('\t')
                datasetList.append(lineList)
            f.close()
        except IOError, (errno, strerror):
            print ("Could not Read File!")
            print ("I/O error(%s): %s" % (errno, strerror))
            raise
        except ValueError:
            print ("Could not convert data to an integer.")
            raise
        except:
            print ("Unexpected error:", sys.exc_info()[0])
            raise
        #*************************************************************************************************
        return datasetList
    
    
    def characterizeDataset(self, rawTrainData):
        " Detect basic dataset parameters "
        #Detect Instance ID's and save location if they occur.
        if cons.labelInstanceID in self.trainHeaderList:
            self.areInstanceIDs = True
            self.instanceIDRef = self.trainHeaderList.index(cons.labelInstanceID)
            print "DataManagement: Instance ID Column location = "+str(self.instanceIDRef)
            self.numAttributes = len(self.trainHeaderList)-2 #one column for InstanceID and another for the Class.
        else:
            self.numAttributes = len(self.trainHeaderList)-1
        
        if cons.labelClass in self.trainHeaderList:
            self.classRef = self.trainHeaderList.index('Class')
            print "DataManagement: Class Column location = "+str(self.classRef)
        else:
            print "DataManagement: Error - Class column not found!  Check data set to ensure correct class column label, or inclusion in the data."

        if self.areInstanceIDs:
            if self.classRef > self.instanceIDRef:
                self.trainHeaderList.pop(self.classRef)
                self.trainHeaderList.pop(self.instanceIDRef)
            else:
                self.trainHeaderList.pop(self.instanceIDRef)
                self.trainHeaderList.pop(self.classRef)
        else:
            self.trainHeaderList.pop(self.classRef)
            
        self.numTrainInstances = len(rawTrainData)
        print "DataManagement: Number of Attributes = " + str(self.numAttributes) #DEBUG
        print "DataManagement: Number of Instances = " + str(self.numTrainInstances) #DEBUG


    def compareDataset(self, rawTestData):
        " Ensures that key dataset parameters are indeed the same for training and testing datasets "
        if self.areInstanceIDs:
            if self.classRef > self.instanceIDRef:
                self.testHeaderList.pop(self.classRef)
                self.testHeaderList.pop(self.instanceIDRef)
            else:
                self.testHeaderList.pop(self.instanceIDRef)
                self.testHeaderList.pop(self.classRef)
        else:
            self.testHeaderList.pop(self.classRef)
            
        if self.trainHeaderList != self.testHeaderList:
            print "DataManagement: Error - Training and Testing Dataset Headers are not equivalent"

        self.numTestInstances = len(rawTestData)
        print "DataManagement: Number of Attributes = " + str(self.numAttributes) #DEBUG
        print "DataManagement: Number of Instances = " + str(self.numTestInstances) #DEBUG
    
    
    def discriminateClasses(self, rawData):
        """  """
        print "DataManagement: Detecting Classes..."
        inst = 0
        classCount = {}
        while inst < self.numTrainInstances:
            target = rawData[inst][self.classRef]
            if target in self.classList:
                classCount[target] += 1 #NOTE: Could potentially store state frequency information to guide learning.
            else:
                self.classList.append(target)
                classCount[target] = 1
            inst += 1
        print "DataManagement: Following Classes Detected:"
        print self.classList
        for each in classCount.keys():
            print "Class: "+str(each)+ " count = "+ str(classCount[each])


    def discriminateAttributes(self, rawData):
        """ Determine whether attributes are Discrete or Continuous """
        print "DataManagement: Detecting Attributes..."
        discreteCount = 0
        continuousCount = 0
        for att in range(len(rawData[0])):
            if att != self.instanceIDRef and att != self.classRef:  #Get just the attribute columns (ignores class and instanceID columns)
                attIsDiscrete = True
                inst = 0
                stateDict = {}
                while attIsDiscrete and len(stateDict.keys()) <= cons.discreteAttributeLimit and inst < self.numTrainInstances:  #Checks which discriminate between discrete and continuous attribute
                    target = rawData[inst][att]
                    if target in stateDict.keys():  #Check if we've seen this attribute state yet.
                        stateDict[target] += 1
                    elif target == cons.labelMissingData: #Ignore missing data
                        pass
                    else: #New state observed
                        stateDict[target] = 1
                    inst += 1

                if len(stateDict.keys()) > cons.discreteAttributeLimit:
                    attIsDiscrete = False
                if attIsDiscrete:
                    self.attributeInfo.append([0,[]])    
                    discreteCount += 1
                else:
                    self.attributeInfo.append([1,[float(target),float(target)]])   #[min,max]
                    continuousCount += 1
        print "DataManagement: Identified "+str(discreteCount)+" discrete and "+str(continuousCount)+" continuous attributes." #Debug

            
    def characterizeAttributes(self, rawData):
        """ Determine range or states of each attribute.  Requires both training and testing datasets in order to standardize formatting across both. """
        print "DataManagement: Characterizing Attributes..."
        attributeID = 0
        for att in range(len(rawData[0])):
            if att != self.instanceIDRef and att != self.classRef:  #Get just the attribute columns (ignores class and instanceID columns)
                for inst in range(len(rawData)):
                    target = rawData[inst][att]
                    if not self.attributeInfo[attributeID][0]: #If attribute is discrete
                        if target in self.attributeInfo[attributeID][1] or target == cons.labelMissingData:
                            pass  #NOTE: Could potentially store state frequency information to guide learning.
                        else:
                            self.attributeInfo[attributeID][1].append(target)
                    else: #If attribute is continuous
                        #Find Minimum and Maximum values for the continuous attribute so we know the range.
                        if target == cons.labelMissingData:
                            pass
                        elif float(target) > self.attributeInfo[attributeID][1][1]:  #error
                            self.attributeInfo[attributeID][1][1] = float(target)
                        elif float(target) < self.attributeInfo[attributeID][1][0]:
                            self.attributeInfo[attributeID][1][0] = float(target)
                        else:
                            pass
                attributeID += 1
            
            
    def formatData(self,rawData):
        """ Get the data into a format convenient for the algorithm to interact with. Our format is consistent with our rule representation, namely, Attribute-list knowledge representation (ALKR),"""
        formatted = []
        #Initialize data format---------------------------------------------------------
        for i in range(len(rawData)):  
            formatted.append([None,None,None]) #[Attribute States, Class, InstanceID]

        for inst in range(len(rawData)):
            stateList = []
            attributeID = 0
            for att in range(len(rawData[0])):
                if att != self.instanceIDRef and att != self.classRef:  #Get just the attribute columns (ignores class and instanceID columns)
                    target = rawData[inst][att]
                    
                    if self.attributeInfo[attributeID][0]: #If the attribute is continuous
                        if target == cons.labelMissingData:
                            stateList.append(target) #Missing data saved as text label
                        else:
                            stateList.append(float(target)) #Save continuous data as floats.
                            
                    else: #If the attribute is discrete - Format the data to correspond to the GABIL (DeJong 1991)
                        tempList = []
                        if target == cons.labelMissingData:
                            tempList = target #Missing data saved as text label
                        else: 
                            for i in range(len(self.attributeInfo[attributeID][1])): #For each possible state of the attribute
                                if target == self.attributeInfo[attributeID][1][i]:  #Find the correct state in the list of possible states, and encode the data in the GABIL manner.  (i.e. [01],[001] = Attribute 1 has state 2, and Attribute 2 has state 1.  Missing Data is encoded as all zero's.
                                    tempList.append(1)
                                else:
                                    tempList.append(0)
          
                        stateList.append(tempList) #Save discrete data in the GABIL format for faster matching.
                    attributeID += 1
                    
            #Final Format-----------------------------------------------
            formatted[inst][0] = stateList                           #Attribute states stored here
            formatted[inst][1] = rawData[inst][self.classRef]        #Class stored here
            if self.areInstanceIDs:
                formatted[inst][2] = rawData[inst][self.instanceIDRef]   #Instance ID stored here
            #-----------------------------------------------------------

        shuffle(formatted) #One time randomization of the order the of the instances in the data, so that if the data was ordered by class, this potential learning bias (based on instance ordering) is eliminated.     
        return formatted