"""
Name: EK_AF_UCS_Environment.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA
Contact: ryan.j.urbanowicz@darmouth.edu
Created: Sept. 4, 2012
Description:  In the context of data mining and classification tasks, the environment is a data set with a limited number of instances with X attributes 
            and some endpoint (typically a discrete class) of interest.  This module loads the data set, automatically detects features of the data.
---------------------------------------------------------------------------------------------------------------------------------------------------------
EK_AF_UCS Version 2.0: An expanded learning classifier system algorithm implemented with our new rule compaction strategies for data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

from EK_AF_UCS_DataManagement import DataManagement
from EK_AF_UCS_Constants import * #Constants module referenced by 'cons'
from EK_AF_UCS_RAIN import *
import sys

class Environment:
    def __init__(self):
        """  """
        #Initialize global variables-------------------------------------------------
        self.dataRef = 0
        self.isEvaluation = False  #Are we performing an evaluation of the whole rule population on the training data (i.e. not a learning iteration)
        self.storeDataRef = 0
        
        self.formatData = DataManagement(cons.trainFile, cons.testFile)
        
        self.currentTrainState = self.formatData.trainFormatted[self.dataRef][0]
        self.currentTrainClass = self.formatData.trainFormatted[self.dataRef][1]
        if cons.testFile == 'None':
            pass
        else:
            self.currentTestState = self.formatData.testFormatted[self.dataRef][0]
            self.currentTestClass = self.formatData.testFormatted[self.dataRef][1]
        

    def getTrainInstance(self):
        """ Returns the current training instance. """  #NOTE: Probably faster way of doing this than additional 'if' statement every learning iteration
        if cons.doArtificialNoise and not self.isEvaluation and cons.firstEpochComplete: #We don't want to add noise to the training data when performing a complete evaluation of the rule population.
            newTrainState = rain.makeItRAIN(self.currentTrainState, self.formatData)
            return [newTrainState, self.currentTrainClass] #Return training instance which has been probabilistically permuted (artificial noise added to the data)
        else:
            return [self.currentTrainState, self.currentTrainClass] #Return unadulterated training data
        
        
    def getTestInstance(self):
        """ Returns the current training instance. """
        return [self.currentTestState, self.currentTestClass]
    
    
    def newInstance(self, isTraining): 
        """  Shifts the environment to the next instance in the data. """
        if isTraining: #Training Data
            if self.dataRef < (self.formatData.numTrainInstances-1):
                self.dataRef += 1
                self.currentTrainState = self.formatData.trainFormatted[self.dataRef][0]
                self.currentTrainClass = self.formatData.trainFormatted[self.dataRef][1]

            else: #Once learning has completed an epoch (i.e. learning iterations though the entire dataset) it starts back at the first instance in the data)
                self.resetDataRef(isTraining)

        else: #Testing Data
            if self.dataRef < (self.formatData.numTestInstances-1):
                self.dataRef += 1
                self.currentTestState = self.formatData.testFormatted[self.dataRef][0]
                self.currentTestClass = self.formatData.testFormatted[self.dataRef][1]
      
      
    def resetDataRef(self, isTraining):
        """ Resets the iteration count through the current data set. """
        self.dataRef = 0 
        if isTraining:
            self.currentTrainState = self.formatData.trainFormatted[self.dataRef][0]
            self.currentTrainClass = self.formatData.trainFormatted[self.dataRef][1]
        else:
            self.currentTestState = self.formatData.testFormatted[self.dataRef][0]
            self.currentTestClass = self.formatData.testFormatted[self.dataRef][1]


    def startEvaluationMode(self):
        """ Turns on evaluation mode.  Saves the instance we left off in the training data. Also important when using RAIN."""
        self.isEvaluation = True
        self.storeDataRef = self.dataRef
        
        
    def stopEvaluationMode(self):
        """ Turns off evaluation mode.  Re-establishes place in dataset."""
        self.isEvaluation = False
        self.dataRef = self.storeDataRef
