"""
Name: EK_AF_UCS_ExpertKnowledge.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA
Contact: ryan.j.urbanowicz@darmouth.edu
Created: Sept. 4, 2012
---------------------------------------------------------------------------------------------------------------------------------------------------------
EK_AF_UCS Version 2.0: An expanded learning classifier system algorithm implemented with our new rule compaction strategies for data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

import math

class ExpertKnowledge:
    def __init__(self, EK_source, EK_max, EK_dop, p_gen):
        """ Expert knowledge initialization - This method may need to be modified 
        depending on the format of your expert knowledge source file."""
        #EXTRACT EK from source file--------------------------------------------
        from EK_AF_UCS_Constants import *
        self.weightList = []
        try:
            fileObject = open(EK_source, 'r')  # opens each datafile to read.
        except:
            print "ExpertKnowledge: Error - EK source file not found!"
            
        tempLine = None
        for i in range(3): #gets through garbage
            tempLine = fileObject.readline()
            
        self.headerDict = {}
        valueList = []
        for i in range(cons.env.formatData.numAttributes):
            tempLine = fileObject.readline()
            tempList = tempLine.strip().split('\t')
            self.headerDict[tempList[0]] = float(tempList[1]) #all attributes and corresponding EK scores are hashed in a dictionary
            valueList.append(float(tempList[1]))
        #-----------------------------------------------------------------------
        #Order the list by SNP ID
        self.orderedValueList = []
        for i in range(cons.env.formatData.numAttributes):
            self.orderedValueList.append(float(self.headerDict[cons.env.formatData.trainHeaderList[i]]))
        print "Ordered Expert Knowledge:"
        print str(self.orderedValueList)
        self.EKweights = self.logisticAttWeights(self.orderedValueList, EK_max, EK_dop, p_gen)
        print "Expert Knowledge Probabilities:"
        print self.EKweights


    def logisticAttWeights(self, rawDataList, EK_max, EK_dop, p_gen):
        """ Uses the logistic function to normalize the weights. """
        print "Calculating alpha and beta for logistic function transformation ..."
        C = p_gen * len(rawDataList) #C will be the sum of the resulting probabilities list, or the number of attributes that should be selected by chance

        #Step 1: Compute the range-------------------------------------------
        minEK = min(rawDataList)
        maxEK = max(rawDataList)
        rangeEK = maxEK - minEK

        #Step 2: Shift data to non-negative----------------------------------
        if minEK < 0:
            for i in range(len(rawDataList)):
                rawDataList[i] = rawDataList[i] - minEK
            maxEK = maxEK - minEK

        #Step 3: Calculate beta----------------------------------------------
        beta = 2*math.log((1-(0.5-EK_max))/(0.5-EK_max))/rangeEK
        print "beta = " +str(beta)

        #Step 4: Calculate alpha via Newton-Raphson--------------------------
        alpha_calc_failed = False
        alpha = -beta*(rangeEK/2) #initial guess
        old_alpha = alpha - 1.0
        while math.fabs(alpha - old_alpha) > pow(10,-EK_dop) and not alpha_calc_failed: #while our guess isn't close enough
            old_alpha = alpha
            try:
                alpha = alpha - ((self.logitsum(rawDataList,alpha,beta)-C)/(self.logitprimesum(rawDataList,alpha,beta)))
            except:
                alpha_calc_failed = True
                print("ExpertKnowledge: Error - failed to compute alpha.")
                break
        print "alpha = " +str(alpha)

        normalizedData = self.logitFinal(rawDataList, alpha, beta) #Apply logistic function to raw EK scores to get probabilities
        return normalizedData
    
    
    def logitsum(self, rawDataList, a, b):
        """ compute the sum of the probabilities. """
        sum = 0.0
        for i in range(len(rawDataList)):
            sum += self.logit(rawDataList[i],a,b)
        return sum
    
    
    def logit(self, rawData, a, b):
        """ Do logistic regression on a datum """
        return 1/(1 + math.exp(-(a+b*rawData)))
    
    
    def logitprimesum(self, rawDataList, a, b):
        """  Compute the derivative of the sum of the probabilities. """
        sum = 0.0
        for i in range(len(rawDataList)):
            sum += self.logitprime(rawDataList[i],a,b)
        return sum
       
        
    def logitprime(self, rawData, a, b):
        """ Take the derivative of logit on a datum. """
        return self.logit(rawData,a,b)*(1-self.logit(rawData,a,b))
    

    def logitFinal(self, rawDataList, a, b):
        """ Do logistic regression on the data. """
        selectionP = []
        for i in range(len(rawDataList)):
            selectionP.append(self.logit(rawDataList[i],a,b))
        return selectionP
    
    def getEKWeights(self):
        """ Returns EKweights. """
        return self.EKweights

