"""
Name: EK_AF_UCS_Main.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA
Contact: ryan.j.urbanowicz@darmouth.edu
Created: Sept. 4, 2012
Description: This module is called to run EK_AF_UCS script from command line.
            It calls all of the overhead required to run EK_AF_UCS including data and parameter management,
            the timer, and the EK_AF_UCS algorithm itself. 
Additional Notes: EK_AF_UCS stands for "Expert Knowledge and Attribute Feedback Supervised Classifying System".
            EK_AF_UCS is a Michigan-Style Learning Classifier System (LCS) based largely on XCS and more precisely on UCS.
            This algorithm is intended for online supervised learning within a finite data mining/classification problem (i.e. single-step).
            Attributes may be either discrete or continuous, but the endpoint/class(s) must be discrete.
            This algorithm would NOT be appropriate for function approximation, behavioral modeling, or other multi-step problems.
---------------------------------------------------------------------------------------------------------------------------------------------------------
EK_AF_UCS Version 2.0: An expanded learning classifier system algorithm implemented with our new rule compaction strategies for data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

#Import Required Modules
import sys
from EK_AF_UCS_Timer import Timer
from EK_AF_UCS_ConfigParser import ConfigParser
from EK_AF_UCS_Environment import Environment
from EK_AF_UCS_Algorithm import EK_AF_UCS
from EK_AF_UCS_Constants import *
from EK_AF_UCS_AttributeTracking import AttributeTracking

helpstr = """
Failed attempt to run EK_AF_UCS.  
EK_AF_UCS requires a specifically formatted configuration file to run (default example included with code).
Please include the file path to the configuration file as an argument to EK_AF_UCS.
"""

#try:
    #Obtain path to configuration file
configurationFile = sys.argv[1] 
    
    #Initialize the Parameters object - this will parse the configuration file and store all constants and parameters.
ConfigParser(configurationFile)
    
    #Engage Timer - tracks run time of algorithm and it's components.
timer = Timer() #TIME
cons.referenceTimer(timer)
    
    #Initialize the Environment object - this manages the data presented to EK_AF_UCS 
env = Environment()
cons.referenceEnv(env) #Send reference to environment object to constants - to access from anywhere in EK_AF_UCS
cons.parseIterations()
     
    #Instantiate EK_AF_UCS Algorithm
algorithm = EK_AF_UCS()

if cons.onlyRC:
        algorithm.runRConly()
else: #Run the EK_AF_UCS algorithm.
    if cons.useExpertKnowledge: #Transform EK scores into probabilities weights for covering. Done once. EK must be externally provided.
        cons.makeExpert() #stores the calculated EK probabilities in EK_AF_UCS_Constants.
            
    if cons.doAttributeTracking:
        cons.timer.startTimeAT()
        AT = AttributeTracking(True)
        cons.timer.stopTimeAT()
    else:
        AT = AttributeTracking(False)
    cons.referenceAttributeTracking(AT)
        
    algorithm.runEK_AF_UCS()

#except:
#   print (helpstr)
#    sys.exit()
    
#MAJOR MODIFICATIONS TO EK_AF_UCS:
#Cleaner organization of the code. Better annotated.
#New Data Management Class
#New rule representation to speed up matching and learning.
#Added ability to handle any attribute type, continuous, discrete, any range or number of alleles
#Altered set construction such that copies of existing classifiers are not made, instead references are used to save memory.

    
    
#Continuing Development -------------------------------------------------------------------------------------------------------------------
#How attribute tracking is calculated, and how the weight probabilities are determined.
#is there value in checking that I made a change to a rule with the GA, do i just check later before adding the rule to the population whether to increase numerosity? or do i do this in subsumption?
#crossover mech - att in both rules, combine instead?  because GA works in correct set
#future considereations - what if training data doesn't have all classes - backup to make guess and report an unseen class or attribute allele.
#Deletion - consider changing how the voting calculation is completed, and when a rule is eligible for deletion.
#where or how is correct set size used
#interaction between correct set size, deletion
#NOTE:::::important, for reboot, and storing pop - have to save the atttribute key somewhere so we know the attribute state order for [001], etc.  save after attribute name (['1','2','0']
#Have a module dedicated to printing the population perhaps - have to translate the encoding to a readable output, for user and for reboot.
#go back over all reboot methods (make sure references are correct or set up some sort of dictionary approach) - dictionary in Classifier
#Double check selection methods
#IDEA - run GA in correct set, unless it has only one rule, then run in match set. 
#look into adaptive parameters
#Before evaluation - perhaps delete completely general rule (if it exists in population) - so we get accurate metrics for covering.
#HAVE to go back and finish off the RAIN stuff - get rule gen, check missing methods.
#make it so that after reboot, the algorithm starts at the same data ref point.
#accuracy calculation - reconsider old way of handling nomatch - add a uniform half right half wrong to maintain consistency in rule compaction, and keep prediction fair. (not penalized too harshly or too little)
#need to specifiy spot to calculate gensum for EK
#Adaptive parameters could be recalculated once an epoch.
#once we have better fitness metric use it instead of accuracy or specificity to guide RAIN.
#TEST CHECKS TO COMPLETE:
#data with missing data
#test all standard algorithm run parameters, on /off etc.
#anywhere that accuracy is calculated - now use getaccuracy method in classifier
#fix average generality, or maybe switch it to average specificity
#Deletion not working right - pop size still growing
#why does rule pop seem to start out with > than 0.5 general
#output number of each class in initial summary - also give regular accuracy compared to balanced.  For balanced data this should be the same - check
#fitness metric idea - covering vs. generality, low gen low cov = low fit, high gen, low cove = high fit.
#major accuracy failure - no rules are getting correct count. - should be getting at least 1 due to covering.
#balance accuracy calculation - problem with calculation in training data
#fix stout summaries so lines  are same length (austhetics
#report accuracy stout, included non-balanced accuracy
#Deletion problem - as soon as rules experience epoch, they get deleted because there are so many rules that haven't finished epoch
#fix - make the pop size limit only include rules that have reached epoch. - This can be a way of automatically adjusting pop size. to handle problem

#START HERE!!!!!!!!!!!!
#Att track finish - and output files
#Debug!!!