"""
Name: EK_AF_UCS_OutputFileManager.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA
Contact: ryan.j.urbanowicz@darmouth.edu
Created: Sept. 4, 2012
Description: This module contains the methods for generating the different output files generated by EK_AF_UCS.  
            These files are generated at each learning checkpoint, and the last iteration.  These include...
            writePopStats:   Summary of the population statistics
            writePop:        Outputs a snapshot of the entire rule population including classifier conditions, classes, and parameters.
            attCo_Occurence: Calculates and outputs co-occurrence scores for each attribute pair in the dataset.
---------------------------------------------------------------------------------------------------------------------------------------------------------
EK_AF_UCS Version 2.0: An expanded learning classifier system algorithm implemented with our new rule compaction strategies for data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

from EK_AF_UCS_Constants import *
from EK_AF_UCS_AttributeTracking import *
import copy

class OutputFileManager:     
    ######################################################################################################################################################################  
    def writePopStats(self, outFile, trainEval, testEval, exploreIter, pop, correct):
        """ Makes output text file which includes all of the parameter settings used in the run as well as all of the evaluation stats including Time Track Output. """
        try:  
            popStatsOut = open(outFile + '_'+ str(exploreIter)+'_PopStats.txt','w') # Outputs Population run stats
            print "Writing Population Statistical Summary File..."
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise
        
        #Evaluation of pop
        popStatsOut.write("Performance Statistics:------------------------------------------------------------------------\n")
        popStatsOut.write("Training Accuracy\tTesting Accuracy\tTraining Coverage\tTesting Coverage\n")
        popStatsOut.write(str(trainEval[0])+"\t")
        if cons.testFile != 'None':
            popStatsOut.write(str(testEval[0])+"\t")
            popStatsOut.write(str(trainEval[1]) +"\t")
            popStatsOut.write(str(testEval[1])+"\n\n") 
        else:
            popStatsOut.write("NA\t")
            popStatsOut.write(str(trainEval[1]) +"\t")
            popStatsOut.write("NA\n\n")
        
        popStatsOut.write("Population Characterization:------------------------------------------------------------------------\n")  
        popStatsOut.write("MacroPopSize\tMicroPopSize\tGenerality\n") 
        popStatsOut.write(str(len(pop.popSet))+"\t"+ str(pop.numerositySum)+"\t"+str(pop.aveGenerality)+"\n\n")

        popStatsOut.write("SpecificitySum:------------------------------------------------------------------------\n")       
        headList = cons.env.formatData.trainHeaderList #preserve order of original dataset

        for i in range(len(headList)): 
            if i < len(headList)-1:
                popStatsOut.write(str(headList[i])+"\t")
            else:
                popStatsOut.write(str(headList[i])+"\n")
                
        # Prints out the Specification Sum for each attribute 
        for i in range(len(pop.attributeSpecList)):
            if i < len(pop.attributeSpecList)-1:
                popStatsOut.write(str(pop.attributeSpecList[i])+"\t")
            else:
                popStatsOut.write(str(pop.attributeSpecList[i])+"\n")  
                      
        popStatsOut.write("\nAccuracySum:------------------------------------------------------------------------\n")
        for i in range(len(headList)): 
            if i < len(headList)-1:
                popStatsOut.write(str(headList[i])+"\t")
            else:
                popStatsOut.write(str(headList[i])+"\n")
                
        # Prints out the Accuracy Weighted Specification Count for each attribute 
        for i in range(len(pop.attributeAccList)):
            if i < len(pop.attributeAccList)-1:
                popStatsOut.write(str(pop.attributeAccList[i])+"\t")
            else:
                popStatsOut.write(str(pop.attributeAccList[i])+"\n")    

        if cons.onlyRC: # When RC ONLY, there is no AttributeTrackingGlobalSums
            pass
        elif cons.doAttributeTracking:
            popStatsOut.write("\nAttributeTrackingGlobalSums:------------------------------------------------------------------------\n")
            for i in range(len(headList)): 
                if i < len(headList)-1:
                    popStatsOut.write(str(headList[i])+"\t")
                else:
                    popStatsOut.write(str(headList[i])+"\n")
                    
            sumGlobalAttTrack = cons.AT.sumGlobalAttTrack()        
            for i in range(len(sumGlobalAttTrack)):
                if i < len(sumGlobalAttTrack)-1:
                    popStatsOut.write(str(sumGlobalAttTrack[i])+"\t")
                else:
                    popStatsOut.write(str(sumGlobalAttTrack[i])+"\n") 
        else:
            popStatsOut.write("\nAttributeTrackingGlobalSums:----Tracking not applied!-----------------------------------------------\n")
            for i in range(len(headList)): 
                if i < len(headList)-1:
                    popStatsOut.write(str(headList[i])+"\t")
                else:
                    popStatsOut.write(str(headList[i])+"\n")
            for i in range(len(headList)): 
                if i < len(headList)-1:
                    popStatsOut.write(str(0.0)+"\t")
                else:
                    popStatsOut.write(str(0.0)+"\n")
                  
        #Time Track ---------------------------------------------------------------------------------------------------------     
        popStatsOut.write("\nRun Time(in minutes):------------------------------------------------------------------------\n") 
        popStatsOut.write(cons.timer.reportTimes())
        popStatsOut.write("\nCorrectTrackerSave:------------------------------------------------------------------------\n")
        for i in range(len(correct)):
            popStatsOut.write(str(correct[i])+"\t")
                
        try:
            popStatsOut.close()
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise   
        
        
    ######################################################################################################################################################################      
    def writePop(self, outFile, exploreIter, pop):
        """ Writes a tab delimited text file specifying the evolved rule population, including conditions, classes, and all rule parameters. """
        try:  
            rulePopOut = open(outFile + '_'+ str(exploreIter)+'_RulePop.txt','w') # Outputs tab delimited text file of rule population and respective rule stats
            print "Writing Population as Data File..."
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise
        #Write Header-----------------------------------------------------------------------------------------------------------------------------------------------
        dataLink = cons.env.formatData
        headList = dataLink.trainHeaderList
        for i in range(len(headList)):   
            rulePopOut.write(str(headList[i])+"\t")
        rulePopOut.write("Class\tFitness\tAccuracy\tCorrectTrack\tMatchTrack\tNumerosity\tSpecificity\tLastCorrectSetSize\tTimeStampGA\tInitTimeStamp\tEpochComplete\tDeletionProb\n")
        #-----------------------------------------------------------------------------------------------------------------------------------------------------------
        #Print whether an attribute is discrete or continuous
        for i in range(len(dataLink.attributeInfo)):
            if dataLink.attributeInfo[i][0]: #If continuous attribute
                if i < len(dataLink.attributeInfo)-1:
                    rulePopOut.write(str(1)+"\t")
                else:
                    rulePopOut.write(str(1)+"\n")   
            else:
                if i < len(dataLink.attributeInfo)-1:
                    rulePopOut.write(str(0)+"\t")
                else:
                    rulePopOut.write(str(0)+"\n")  
        
        #Print key for state order of discrete variables, and range of variable for continuous.
        for i in range(len(dataLink.attributeInfo)):
            if dataLink.attributeInfo[i][0]: #If continuous attribute
                if i < len(dataLink.attributeInfo)-1:
                    rulePopOut.write(str(dataLink.attributeInfo[i][1][0])+str(cons.alleleSeparation)+str(dataLink.attributeInfo[i][1][1])+"\t")
                else:
                    rulePopOut.write(str(dataLink.attributeInfo[i][1][0])+str(cons.alleleSeparation)+str(dataLink.attributeInfo[i][1][1])+"\n")
            else: #If discrete attribute
                if i < len(dataLink.attributeInfo)-1:
                    alleleString = ''
                    j = 0
                    for allele in dataLink.attributeInfo[i][1]:
                        if j < len(dataLink.attributeInfo[i][1])-1:
                            alleleString += str(allele)+str(cons.alleleSeparation)
                        else:
                            alleleString += str(allele)
                        j += 1
                    rulePopOut.write(str(alleleString)+"\t")
                else:
                    alleleString = ''
                    j = 0
                    for allele in dataLink.attributeInfo[i][1]:
                        if j < len(dataLink.attributeInfo[i][1])-1:
                            alleleString += str(allele)+str(cons.alleleSeparation)
                        else:
                            alleleString += str(allele)
                        j += 1
                    rulePopOut.write(str(alleleString)+"\n")
        #Write each classifier--------------------------------------------------------------------------------------------------------------------------------------
        for cl in pop.popSet:
            rulePopOut.write(str(cl.printClassifier())) 
        #-----------------------------------------------------------------------------------------------------------------------------------------------------------
        
        try:
            rulePopOut.close() 
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise   
        

    ######################################################################################################################################################################    
    def attCo_Occurence(self, outFile, exploreIter, pop):
        """ Calculates pairwise attribute co-occurence througout all rules in the population."""
        print "Calculating Attribute Co-occurence Scores..."
        dataLink = cons.env.formatData
        comboList = []
        castList = [None,None,0,0] #att1, att2, Specificity, Accuracy Weighted Specificity
        count = 0
        dim = dataLink.numAttributes
        #Specify all attribute pairs.
        for i in range(0, dim-1):
            for j in range(i+1,dim):
                comboList.append(copy.deepcopy(castList))
                comboList[count][0] = dataLink.trainHeaderList[i]
                comboList[count][1] = dataLink.trainHeaderList[j]   
                count += 1

        for cl in pop.popSet:
            count = 0
            for i in range(dim-1):
                for j in range(i+1,dim):
                    if i in cl.specifiedAttList and j in cl.specifiedAttList:
                        comboList[count][2] += cl.numerosity
                        comboList[count][3] += cl.numerosity * cl.getAccuracy()
                    count += 1
        
        tupleList = []
        for i in comboList:
            tupleList.append((i[0],i[1],i[2],i[3]))
        sortedComboList = sorted(tupleList,key=lambda test: test[3], reverse=True)
        print "Writing Attribute Co-occurence scores as data file..."
        f = open(outFile + '_'+ str(exploreIter)+ '_CO.txt', 'w')
        for i in range(len(sortedComboList)):
            for j in range(len(sortedComboList[0])): #att1, att2, count, AWcount
                if j < len(sortedComboList[0])-1:
                    f.write(str(sortedComboList[i][j])+'\t')
                else:
                    f.write(str(sortedComboList[i][j])+'\n')
        f.close()
        
        
    ######################################################################################################################################################################    
    def save_tracking(self, exploreIter, outFile):
        """ Prints out Attribute Tracking scores to txt file. """    
        try:  
            f = open(outFile + '_'+ str(exploreIter + 1)+'_AttTrack.txt','w') # Outputs tab delimited text file of rule population and respective rule stats
            print "Writing Attribute Tracking as Data File..."
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise
        
        if cons.adjustedAT:
            trackingSums = cons.AT.attAccuracySumsAdj
        else:
            trackingSums = cons.AT.attAccuracySums
            
        #-------------------------------------------------------------------
        if cons.env.formatData.areInstanceIDs: 
            f.write(str(cons.labelInstanceID) + '\t') #Write InstanceID label
            for att in cons.env.formatData.trainHeaderList:
                f.write(str(att) + '\t')
            f.write(str(cons.labelClass)+ '\n') #Write class label
            #---------------------------------------------------------------
            for i in range(len(trackingSums)):
                trackList = trackingSums[i]
                f.write(str(cons.env.formatData.trainFormatted[i][2])+ '\t') #Write InstanceID
                for att in trackList:
                    f.write(str(att) + '\t')
                f.write(str(cons.env.formatData.trainFormatted[i][1]) +'\n') #Write Class

        else: #Write attribute tracking without InstanceIDs
            for att in cons.env.formatData.trainHeaderList:
                f.write(str(att) + '\t')
            f.write(str(cons.labelClass)+ '\n') #Write class label
            #---------------------------------------------------------------
            for i in range(len(trackingSums)):
                trackList = trackingSums[i]
                for att in trackList:
                    f.write(str(att) + '\t')
                f.write(str(cons.env.formatData.trainFormatted[i][1]) +'\n') #Write Class
                