"""
Name: EK_AF_UCS_RAIN.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA
Contact: ryan.j.urbanowicz@darmouth.edu
Created: Sept. 4, 2012
Description:  We implement the option of using 'Uniform Targeted RAIN' as described by Urbanowicz 2011.  This is intended to reduce overfitting in noisy problem domains.  
            The user must specify the noise magnitude as a parameter.
---------------------------------------------------------------------------------------------------------------------------------------------------------
EK_AF_UCS Version 2.0: An expanded learning classifier system algorithm implemented with our new rule compaction strategies for data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

from EK_AF_UCS_Constants import * 
from random import *
import copy
import math

class RAIN:
    def makeItRAIN(self, currentTrainState, formattedData): 
        """ Probabilistically adds random noise to attributes in the current training state. 
        Can handle missing data"""

        cons.timer.startTimeRAIN()
        perturbedState = copy.deepcopy(currentTrainState)
        #Determine how many attributes will be perturbed
        maxPermute = 0
        
        id_List = []
        for t in range(len(perturbedState)):
            id_List.append(t) #list with references to attribute positions
            if random() < cons.noiseMagnitude:
                maxPermute += 1

        #-------------------------------------------------------------------
        currentCount = 0
        while currentCount < maxPermute:
            #Adjust weights (such that large weight values become the smallest probability of permutation.
            tempWeights = copy.copy(cons.RAINWeights)
            minW = max(tempWeights)*-1     
            specSum = 0  
            #Calculate the sum
            for ea in id_List:
                tempWeights[ea] = tempWeights[ea] * -1 
                specSum += tempWeights[ea] + math.fabs(minW)
                
            #Roulette Wheel------------------------------------------------
            i=0     #Attribute reference chosen for permutation
            choiceP = random() * specSum
            mysum = tempWeights[id_List[i]] + math.fabs(minW)
            while choiceP > mysum:
                i += 1
                mysum += tempWeights[id_List[i]] + math.fabs(minW)    
            #--------------------------------------------------------------
            #Detect current allele - and remove it (is it discrete or continuous - the change depends on what it is.
            if formattedData.attributeInfo[id_List[i]][0]: #If chosen attribute is continuous
                #Note: we do same thing if there is missing data or not.
                perturbedState[id_List[i]] = uniform(formattedData.attributeInfo[id_List[i]][1][0],formattedData.attributeInfo[id_List[i]][1][1]) #Pick random continuous value within the range observed for this attribute in the training data.
            else: #If chosen attribute is discrete
                if perturbedState[id_List[i]] != cons.labelMissingData: #Change to different allele state
                    alleleID = []
                    for j in range(len(perturbedState[id_List[i]])):
                        if not perturbedState[id_List[i]][j] == 1:
                            alleleID.append(j)                  #Keep locations of potential new alleles
                        else:
                            perturbedState[id_List[i]][j] = 0   #Remove old allele
                    perturbedState[id_List[i]][choice(alleleID)] = 1 #New permuted allele added
                    
                else: #Attribute is missing data - so insert any possible allele.
                    numAlleles = len(formattedData.attributeInfo[id_List[i]][1])
                    allelePosition = randint(0,numAlleles-1)
                    tempList = []
                    for j in range(numAlleles):
                        if j == allelePosition:
                            tempList.append(1)
                        else:
                            tempList.append(0)
                    perturbedState[id_List[i]] = tempList
                    
            currentCount += 1
            id_List.remove(id_List[i]) #ensures that same attribute is not chosen for permutation
        cons.timer.stopTimeRAIN()
        return perturbedState
    
rain = RAIN() 